<?php
/**
 * mail_fetch/setup.php
 *
 * Copyright (c) 1999-2011 CDI (cdi@thewebmasters.net) All Rights Reserved
 * Modified by Philippe Mingo 2001-2009 mingo@rotedic.com
 * An RFC 1939 compliant wrapper class for the POP3 protocol.
 *
 * Licensed under the GNU GPL. For full terms see the file COPYING.
 *
 * POP3 class
 *
 * @copyright 1999-2011 The SquirrelMail Project Team
 * @license http://opensource.org/licenses/gpl-license.php GNU Public License
 * @package plugins
 * @subpackage mail_fetch
 */

class POP3 {
    var $ERROR      = '';       //  Error string.

    var $TIMEOUT    = 60;       //  Default timeout before giving up on a
                                //  network operation.

    var $COUNT      = -1;       //  Mailbox msg count

    var $BUFFER     = 512;      //  Socket buffer for socket fgets() calls.
                                //  Per RFC 1939 the returned line a POP3
                                //  server can send is 512 bytes.

    var $FP         = '';       //  The connection to the server's
                                //  file descriptor

    var $MAILSERVER = '';       // Set this to hard code the server name
    var $PORT 		= 110;		// Server port

    var $DEBUG      = FALSE;    // set to true to echo pop3
                                // commands and responses to error_log
                                // this WILL log passwords!

    var $BANNER     = '';       //  Holds the banner returned by the
                                //  pop server - used for apop()

    var $ALLOWAPOP  = FALSE;    //  Allow or disallow apop()
                                //  This must be set to true
                                //  manually
                                
    var $safemode_enabled = true;

    function POP3 ( $server = '', $timeout = 0, $port = 110, $connect = false ) {

        $this->safemode_enabled = (bool)ini_get('safe_mode');
        $this->BUFFER 			= (int)$this->BUFFER;
        
        $success = $this->_init( $server, $port, $timeout );
        
        if ($connect && $success) 
        	$this->connect( $this->MAILSERVER, $this->PORT );
    }
    
    private function _init($server = '', $port = 110, $timeout = 0){

    	// Do not allow programs to alter MAILSERVER
        // if it is already specified. They can get around
        // this if they -really- want to, so don't count on it.
    	if (empty($this->MAILSERVER)) {
    		if (!empty($server))
    			$this->MAILSERVER = $server;
    	}
    	
    	if ($timeout) {
    		$this->TIMEOUT = (int)$timeout;
    		if (!$this->safemode_enabled) 
    			set_time_limit( $this->TIMEOUT );
    	}
    	
    	$port = (int)$port;
    	if ($port) {
    		$this->PORT = $port;
    	}
    	
    	return !empty($this->MAILSERVER);
    }
    
    function set_error( $error , $caller = null, $close = false ) {
    	$this->ERROR = 'POP3 '. ( $caller ? $caller.'()' : '' ).': '. $error;
    	
    	if ($close)
    		unset($this->FP);
    	
    	return false;
    }

    function update_timer () {
        if (!$this->safemode_enabled)
            set_time_limit($this->TIMEOUT);
        return true;
    }

    function connect ($server = '', $port = 110)  {
        
        if ( empty($this->MAILSERVER) || $this->PORT != $port ) {
        	if (!$this->_init( $server, $port )) {
        		return $this->set_error( _("No server specified"), __FUNCTION__, true );
        	}
        }
		
        $this->FP = @fsockopen($this->MAILSERVER, $this->PORT, $errno, $errstr);

        if(!$this->FP) {
            return $this->set_error(  _("Error ") . "[$errno] [$errstr]", __FUNCTION__, true );
        }
        
		function_exists('stream_set_blocking') ? stream_set_blocking($this->FP, 1) : socket_set_blocking($this->FP, -1);

        $this->update_timer();
        $reply = $this->get_reply();
        
        if(!$this->is_ok($reply)) {
            return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__, true );
        }
        
        $this->BANNER = $this->parse_banner($reply);
        return true;
    }
    
     private function send($string) {
    	$raw = $string;
    	
    	if ($this->FP) {
    		$string = strpos($string, "\r\n") === FALSE ? $string."\r\n" : $string;
    		
    		if ( ($number = @fwrite($this->FP, $string)) ) {
    			if ($this->DEBUG) {
    				@error_log('POP3 '. __FUNCTION__.'(): ['.$raw.']');
    			}
    			return $number;
    		}
    	}
    	
    	return false;
    }
    
    private function get_reply($strip_clf = true) {
    	$reply = @fgets( $this->FP, $this->BUFFER );
    	    	
    	if ($reply) {
    		if ($strip_clf) {
    			$reply = $this->strip_clf( $reply );
    		}
    		
    		if ($this->DEBUG) {
    			@error_log('POP3 GOT: ['. $reply .'] FROM ['.$this->MAILSERVER.']',0);
    		}
    	}
    	
    	return $reply;
    }

    function user ($user = "") {
        // Sends the USER command, returns true or false

        if( empty($user) )
            return $this->set_error( _("no login ID submitted"), __FUNCTION__ );
        
        if(!isset($this->FP))
            return $this->set_error( _("connection not established"), __FUNCTION__ );
        
        $reply = $this->send_cmd("USER $user");
        if(!$this->is_ok($reply))
        	return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );
        	
        return $this;
    }

    function pass ($pass = "")     {
        // Sends the PASS command, returns # of msgs in mailbox,
        // returns false (undef) on Auth failure

        if(empty($pass))
           return $this->set_error( _("No password submitted"), __FUNCTION__ );
        
        if(!isset($this->FP))
            return $this->set_error( _("connection not established"), __FUNCTION__ );
            
    	$reply = $this->send_cmd("PASS $pass");
        if(!$this->is_ok($reply)) {
            $this->set_error( _("Authentication failed") . " [$reply]", __FUNCTION__ );
            $this->quit();
            return false;
        }
        
        //  Auth successful.
        $this->COUNT = $this->last("count");
        return $this->COUNT;
    }

    function apop ($login,$pass) {
        //  Attempts an APOP login. If this fails, it'll
        //  try a standard login. YOUR SERVER MUST SUPPORT
        //  THE USE OF THE APOP COMMAND!
        //  (apop is optional per rfc1939)

        if(!isset($this->FP))
            return $this->set_error( _("No connection to server"), __FUNCTION__ );
            
        if(!$this->ALLOWAPOP) {
            $retVal = $this->login($login,$pass);
            return $retVal;
        } 
        
        if (empty($login))
            return $this->set_error( _("No login ID submitted"), __FUNCTION__ );
            
        if(empty($pass))
            return $this->set_error( _("No password submitted"), __FUNCTION__ );
            
       
        $banner = $this->BANNER;
        if( !$banner ) {
            $this->set_error(  _("No server banner") . ' - ' . _("abort"), __FUNCTION__ );
            $retVal = $this->login($login,$pass);
            return $retVal;
        }
        
        $AuthString = $banner. $pass;
        $APOPString = md5($AuthString);
        
        $reply = $this->send_cmd("APOP $login $APOPString");
        if(!$this->is_ok($reply)) {
            $this->set_error( _("apop authentication failed") . ' - ' . _("abort"), __FUNCTION__ );
            $retVal = $this->login($login,$pass);
            return $retVal;
        }
        
        //  Auth successful.
        $this->COUNT = $this->last("count");
        return $this->COUNT;
    }

    function login ($login = "", $pass = "") {
        // Sends both user and pass. Returns # of msgs in mailbox or
        // false on failure (or -1, if the error occurs while getting
        // the number of messages.)

        if( !isset($this->FP) ) {
            return $this->set_error( _("No connection to server"), __FUNCTION__ );
        } else {
            $fp = $this->FP;
            if( !$this->user( $login ) ) {
                //  Preserve the error generated by user()
                return false;
            } else {
                $count = $this->pass($pass);
                if( (!$count) || ($count == -1) ) {
                    //  Preserve the error generated by last() and pass()
                    return false;
                } else
                    return $count;
            }
        }
    }

    function top ($msgNum, $numLines = 0) {
        //  Gets the header and first $numLines of the msg body
        //  returns data in an array with each returned line being
        //  an array element. If $numLines is empty, returns
        //  only the header information, and none of the body.

        if(!isset($this->FP))
            return $this->set_error( _("No connection to server"), __FUNCTION__ );
        
        $this->update_timer();

        $this->send( "TOP $msgNum $numLines" );
        $reply = $this->get_reply();
        if (!$this->is_ok($reply)) {
           	return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );
        }

        $MsgArray = array();
        while ( ( $line = fgets($this->FP, $this->BUFFER) ) && !$this->iseof( $line ) ) {
        	$MsgArray[] = $line;
        }

        return $MsgArray;
    }

    function pop_list ($msgNum = '') {
        //  If called with an argument, returns that msgs' size in octets
        //  No argument returns an associative array of undeleted
        //  msg numbers and their sizes in octets

        if (!isset($this->FP))
            return $this->set_error( _("No connection to server"), __FUNCTION__ );
        
        if ($this->COUNT === FALSE || $this->COUNT == -1) {
        	return false;
        }
        
        if ( $this->COUNT == 0 ) {
        	return array(0,0);
        }

        $this->update_timer();
		
        if (!$msgNum) {
           	$this->send( "LIST $msgNum" );

            $reply = $this->get_reply();
            if (!$this->is_ok($reply))
                return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );
                
            // Gmail and some email providers return the word 'messages' in the response, example:
            // +OK 307 messages (23191343 bytes)
            preg_match_all('([0-9]+)', $reply, $m);
            $size = FALSE;
            if( !empty($m[0]) ) {
            	$size = array_pop($m[0]);
            }
            return $size;
        }
        
        $reply = $this->send_cmd('LIST');
        if (!$this->is_ok($reply))
            return $this->set_error( _("Error ") .  "[$reply]", __FUNCTION__ );
            
        $msg = array();    
    	for( $i=1; $i <= $this->COUNT; $i++ ) {
    		$info = $this->get_reply();
    		if ( !empty($info) ) {
    			$info = explode(' ', $info);
    			$msg[ $info[0] ] = $info[1];
    		}
    	}
    	
    	return $msg;
    }

    function get ($msgNum) {
        //  Retrieve the specified msg number. Returns an array
        //  where each line of the msg is an array element.

        if (!isset($this->FP))
        	return $this->set_error( _("No connection to server"), __FUNCTION__ );

        $this->update_timer();
        $reply = $this->send_cmd('RETR '. $msgNum);

        if(!$this->is_ok($reply)) 
        	return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );

        $MsgArray = array();

        while ( ( $line = fgets($this->FP, $this->BUFFER) ) && !$this->iseof( $line ) ) {
        	$MsgArray[] = $line;
        }
        
        return $MsgArray;
    }
    
    private function iseof( $line ) {
    	return $line == ".\r\n" || empty($line);
    }

    function last ( $type = "count" ) {
    	static $saved_last;
        //  Returns the highest msg number in the mailbox.
        //  returns -1 on error, 0+ on success, if type != count
        //  results in a popstat() call (2 element array returned)

        if (!isset($this->FP)) {
        	$this->set_error( _("No connection to server"), __FUNCTION__ );
        	return -1;
        }

        $reply = $this->send_cmd("STAT");
        if(!$this->is_ok($reply)) {
        	$this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );
        	return -1;
        }

        if( !$saved_last ) {
        	list( $ok, $count, $size) = explode( ' ', $reply);
        	$saved_last = array( $count, $size );
        }
        
        return $type == 'count' ? $saved_last['count'] : $saved_last;
    }

    function reset () {
        //  Resets the status of the remote server. This includes
        //  resetting the status of ALL msgs to not be deleted.
        //  This method automatically closes the connection to the server.

        if (!isset($this->FP))
        	return $this->set_error( _("No connection to server"), __FUNCTION__ );
        	
        $reply = $this->send_cmd("RSET");
        if(!$this->is_ok($reply))
        {
            //  The POP3 RSET command -never- gives a -ERR
            //  response - if it ever does, something truely
            //  wild is going on.

            $this->ERROR = "POP3 reset: " . _("Error ") . "[$reply]";
            @error_log("POP3 reset: ERROR [$reply]",0);
        }
        $this->quit();
        return true;
    }

    function send_cmd ( $cmd ) {
        //  Sends a user defined command string to the
        //  POP server and returns the results. Useful for
        //  non-compliant or custom POP servers.

        //  The return value is a standard fgets() call, which
        //  will read up to $this->BUFFER bytes of data, until it
        //  encounters a new line, or EOF, whichever happens first.

        //  This method works best if $cmd responds with only
        //  one line of data.

        if (!isset($this->FP))
            return $this->set_error( _("No connection to server"), __FUNCTION__ );

        if (empty($cmd)) {
            $this->set_error( _("Empty command string") , __FUNCTION__ );
            return "";
        }

        $this->update_timer();
        $reply = false;
        if ( $this->send( $cmd ) ) 
        	$reply = $this->get_reply();
        
        return $reply;
    }

    function quit() {
        //  Closes the connection to the POP3 server, deleting
        //  any msgs marked as deleted.

        if (!isset($this->FP))
            return $this->set_error( _("Connection already closed."), __FUNCTION__ );
        
        $this->send('QUIT') && $this->get_reply();
        
        fclose($this->FP);
        return true;
    }

    function popstat () {
        //  Returns an array of 2 elements. The number of undeleted
        //  msgs in the mailbox, and the size of the mbox in octets.

        $PopArray = $this->last("array");

        if ( empty($PopArray) || $PopArray == -1 )
        	return false;
        
        return $PopArray;
    }

    function uidl ($msgNum = "")
    {
        //  Returns the UIDL of the msg specified. If called with
        //  no arguments, returns an associative array where each
        //  undeleted msg num is a key, and the msg's uidl is the element
        //  Array element 0 will contain the total number of msgs

        if (!isset($this->FP)) 
            return $this->set_error( _("No connection to server"), __FUNCTION__ );

        if (!empty($msgNum)) {
            $reply = $this->send_cmd("UIDL $msgNum");
            
            if(!$this->is_ok($reply))
            	return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );
            
            list ($ok,$num,$myUidl) = preg_split('/\s+/',$reply);
            return $myUidl;
        }
        	
        $this->update_timer();

        $this->send('UIDL');
        $reply = $this->get_reply();
            
    	if ( !$this->is_ok($reply) ) 
            return $this->set_error( _("Error ") . "[$reply]", __FUNCTION__ );

		$response = array();
        while ( ( $line = fgets($this->FP, $this->BUFFER) ) && !$this->iseof( $line ) ) {
        	$exploded = preg_split('#\s+#', $line);
        	$response[ $exploded[0] ] = $exploded[1];
        }
        	
        return $response;
    }

    function delete ($msgNum) {
        //  Flags a specified msg as deleted. The msg will not
        //  be deleted until a quit() method is called.

        if (!isset($this->FP)) 
        	return $this->set_error( _("No connection to server"), __FUNCTION__ );
        	
        if (empty($msgNum))
            return $this->set_error( _("No msg number submitted"), __FUNCTION__ );

        $reply = $this->send_cmd("DELE $msgNum");
        if (!$this->is_ok($reply))
        	return $this->set_error( _("Command failed ") . "[$reply]", __FUNCTION__ );
        
        return true;
    }

    //  *********************************************************

    //  The following methods are internal to the class.

    function is_ok ($cmd) {
        //  Return true or false on +OK or -ERR
        return stristr($cmd, '+OK') !== FALSE;
    }

    function strip_clf ($text) {
        // Strips \r\n from server responses

        return str_replace(array("\r","\n"), '', $text);
    }

    function parse_banner ( $server_text ) {
    	if (empty($server_text)) {
    		return $server_text;
    	}
    	
    	preg_match_all('#<([^<>]+)>#', $server_text, $matches);
    	
    	$banners = array();
    	if (isset($matches[1]) && !empty($matches[1])) {
    		$banners = $matches[1];
    	}
    	
    	return implode('', $banners);
    }
    
    function get_error() {
    	$error  =  $this->ERROR;
    	unset($this->ERROR);
    	return $error;
    }

}   // End class

